static char _senWinEvent_c[] =
"$Header: /cvs/tcl/tclsvc/senWinEvent.c,v 1.3 1999/11/11 17:39:30 matt Exp $";
/*
 * Copyright (C) 1997-1999 Sensus Consulting Ltd.
 * Matt Newman <matt@sensus.org>
 *
 * Adds command win32::eventlog
 *
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <tchar.h>

#include "tcl.h"

/*
 * External Functions
 */

/*
 * Static functions
 */
static int        EventObjCmd _ANSI_ARGS_((ClientData clientData,
                    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));

/*
 * Static Declarations
 */

static int
EventObjCmd (clientdata, interp, objc, objv)
    ClientData clientdata;
    Tcl_Interp *interp;
    int                objc;
    Tcl_Obj        *CONST objv[];
{
    char *def_source = (char*)clientdata;
    HANDLE          hEventSource;
    LPTSTR          lpszStrings[1];
    DWORD        catId = 0, evId = 1;
    char        *opt, *msg, *host = NULL, *source = def_source;
    int                idx, level = EVENTLOG_INFORMATION_TYPE;
    static char *levels[] = {
        "success",
        "error",
        "warning",
        "information",
        "audit/success",
        "audit/failure",
        NULL
    };
    static int idx2lvl[] = {
        EVENTLOG_SUCCESS,
        EVENTLOG_ERROR_TYPE,
        EVENTLOG_WARNING_TYPE,
        EVENTLOG_INFORMATION_TYPE,
        EVENTLOG_AUDIT_SUCCESS,
        EVENTLOG_AUDIT_FAILURE
    };

    for (idx = 1;(objc - idx) > 0;idx += 2) {
        opt = Tcl_GetStringFromObj(objv[idx], (int *) NULL);
        if (opt[0] != '-')
            break;
        if (strcmp( "-host", opt)==0) {
            host = Tcl_GetStringFromObj(objv[idx+1], (int *) NULL);
        } else if (strcmp( "-source", opt)==0) {
            source = Tcl_GetStringFromObj(objv[idx+1], (int *) NULL);
        } else if (strcmp( "-event", opt)==0) {
            if (Tcl_GetIntFromObj(interp, objv[idx+1], &evId) != TCL_OK) {
                return TCL_ERROR;
            }
        } else if (strcmp( "-category", opt)==0) {
            if (Tcl_GetIntFromObj(interp, objv[idx+1], &catId) != TCL_OK) {
                return TCL_ERROR;
            }
        } else if (strcmp( "-level", opt)==0) {
            if (Tcl_GetIndexFromObj(interp, objv[idx+1], levels, "level", 0,
                    &level) != TCL_OK) {
                return TCL_ERROR;
            }
            level = idx2lvl[level];
        } else {
            Tcl_AppendResult(interp, "bad option \"", opt,
                    "\": must be -category, -event, -host, -level, or -source", (char*)NULL);
            return TCL_ERROR;
        }
    }
    switch (objc -idx) {
    case 1:
        msg = Tcl_GetStringFromObj(objv[idx], (int *) NULL);
        break;
    case 2:
        if (Tcl_GetIndexFromObj(interp, objv[idx], levels, "level", 0,
                &level) != TCL_OK) {
            return TCL_ERROR;
        }
        level = idx2lvl[level];
        msg = Tcl_GetStringFromObj(objv[idx+1], (int *) NULL);
        break;
    default:
        Tcl_WrongNumArgs(interp, 1, objv,
                    "?options? ?lvl? message");
        return TCL_ERROR;
    }
    if (*source == '\0') {
        source = def_source;
    }
    hEventSource = RegisterEventSource(TEXT(host), source);
    if (hEventSource == NULL) {
        Tcl_AppendResult(interp, "can't open eventlog for source \"",
                    source, "\"", (char*)NULL);
        return TCL_ERROR;
    }
    lpszStrings[0] = TEXT(msg);
    ReportEvent(hEventSource,        // handle of event source
            (WORD)level,        // event type
            (WORD)catId,        // event category
            evId,                // event ID
            NULL,                // current user's SID
            1,                        // strings in lpszStrings
            0,                        // no bytes of raw data
            lpszStrings,        // array of error strings
            NULL);                // no raw data

    (void) DeregisterEventSource(hEventSource);
    return TCL_OK;
}

int
Senevent_Init(interp)
        Tcl_Interp *interp;
{
    char *def_source;
    char buf[BUFSIZ], *cp;

    cp = (char*)Tcl_GetVar(interp, "tcl_service", TCL_GLOBAL_ONLY);
    if (!cp || !*cp) {
        if (Tcl_Eval(interp, "file tail [info nameofexecutable]")
            == TCL_OK) {
            strcpy( buf, Tcl_GetStringResult(interp));
            if ((cp = strchr(buf, '.')) != NULL)
                *cp = '\0';
            if (buf[0] != '\0') {
                cp = strdup(buf);
            }
        }
    }
    if (cp && *cp)
        def_source = strdup(cp);
    else
        def_source = strdup("Tcl");

    Tcl_ResetResult(interp);

    Tcl_CreateObjCommand(interp, "eventlog", EventObjCmd,
                  (ClientData) def_source, (Tcl_CmdDeleteProc *) NULL);

    return Tcl_PkgProvide(interp, "eventlog", "2.0");
}
